import cv2
import argparse
import numpy as np
import threading
import queue
import time
import copy

# ros2
import rclpy
from rclpy.node import Node
from sensor_msgs.msg import Image
from geometry_msgs.msg import Twist
from cv_bridge import CvBridge
from std_msgs.msg import Int32
from rclpy.executors import MultiThreadedExecutor
from std_srvs.srv import SetBool

# ours
from spacemit_cv import AGVDetection

# 提供控制开启和关闭跟踪的服务
class FollowControlService(Node):
    def __init__(self):
        super().__init__('follow_control_service')
        self.srv = self.create_service(SetBool, 'toggle_follow', self.callback)
        self.ai_enabled = False  # 初始状态
        self.get_logger().info('AI 跟踪控制服务已启动, 当前状态为暂停')
        self.lock = threading.Lock() # 锁，用于线程数据同步

    def callback(self, request, response):
        with self.lock:
            self.ai_enabled = request.data
        if self.ai_enabled:
            msg = 'AI 跟踪模块已开启'
        else:
            msg = 'AI 跟踪模块已关闭'

        self.get_logger().info(f'收到请求: data={request.data} -> {msg}')
        response.success = True
        response.message = msg
        return response

class VelocityPublisher(Node):
    """ROS2 节点：用于发布速度控制指令"""
    def __init__(self):
        super().__init__('velocity_publisher')
        self.publisher = self.create_publisher(Twist, 'cmd_vel', 30)

    def publish_velocity(self, linear_x, angular_z):
        msg = Twist()
        msg.linear.x = linear_x
        msg.angular.z = angular_z
        self.publisher.publish(msg)

class ImagePublisher(Node):
    """ROS2 节点：用于发布处理后的图像"""
    def __init__(self):
        super().__init__('image_publisher')
        self.publisher = self.create_publisher(Image, 'detection_image', 30)
        self.bridge = CvBridge()

    def publish_image(self, frame):
        """发布 ROS2 图像消息"""
        msg = self.bridge.cv2_to_imgmsg(frame, encoding='bgr8')
        self.publisher.publish(msg)

class MyDetectionThread(threading.Thread):
    def __init__(self, result_queue, model_path, image_publisher):
        threading.Thread.__init__(self)
        self.result_queue = result_queue
        self.detector = AGVDetection(model_path)
        self.cap = cv2.VideoCapture(20)  
        # 设置分辨率（宽度和高度）
        self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, 640)  # 设置宽度
        self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 480) # 设置高度
        self.running = True

        self.image_publisher = image_publisher  # 传入 ROS2 图像发布器

    def run(self):
        while self.running:
            ret, frame = self.cap.read()
            # print(frame.shape)
            if ret:
                detections = self.detector.infer(frame)
                if detections:
                    height, width, _ = frame.shape
                    center_x = width // 2
                    center_y = height // 2
                    min_distance = float('inf')
                    closest_box = None
                    for det in detections:
                        x1, y1, x2, y2 = det
                        center = ((x1 + x2) // 2, (y1 + y2) // 2)
                        distance = ((center[0] - center_x) ** 2 + (center[1] - center_y) ** 2) ** 0.5
                        if distance < min_distance:
                            min_distance = distance
                            closest_box = det
                        
                    if closest_box:
                        self.result_queue.put(closest_box)

                        x1, y1, x2, y2 = closest_box
                        cv2.rectangle(frame, (int(x1), int(y1)), (int(x2), int(y2)), (0, 0, 255), 5)
                
                self.image_publisher.publish_image(frame)
            else:
                break

    def stop(self):
        self.running = False
        self.cap.release()

def run_executor(executor):
    executor.spin()  # 运行 ROS 2 事件循环（不会阻塞主线程）

def main():
    rclpy.init()

    # 跟踪开关变量服务器
    follow_service = FollowControlService()

    # 速度发布者
    velocity_publisher = VelocityPublisher()

    # 图像发布者
    image_publisher = ImagePublisher()

    # ros2 多线程管理
    executor = MultiThreadedExecutor()
    executor.add_node(follow_service)

    executor_thread = threading.Thread(target=run_executor, args=(executor,), daemon=True)
    executor_thread.start()
   
    # 目标检测线程
    result_queue = queue.Queue(maxsize=2)
    detection_thread = MyDetectionThread(result_queue, 'spacemit_cv/model/yolov8n.q.onnx', image_publisher)
    detection_thread.start()

    # 主线程，执行跟踪逻辑
    try:
        count = 0
        start_time = time.time()
        closest_box = []
        follow_me_flag = 0
        cmd_x_z = [0.0, 0.0]
        while rclpy.ok():
            
            with follow_service.lock:
                follow_me_flag = follow_service.ai_enabled

            # 更新框
            if not result_queue.empty():
                closest_box = copy.deepcopy(result_queue.get())
                # print(f"最靠近摄像头中心的检测框: {closest_box}")

            # 是否开启跟随
            if not follow_me_flag:
                if abs(cmd_x_z[0]-0.0) >= 0.01 or abs(cmd_x_z[1]-0.0) >= 0.01:
                    cmd_x_z = [0.0, 0.0]
                    for i in range(0, 3):
                        velocity_publisher.publish_velocity(0.0, 0.0)  # 停止小车
                        time.sleep(0.01)
                continue
            
            # 始终跟踪上一次框
            cmd_x_z = [0.0, 0.0]
            if len(closest_box) == 4:
                x1, y1, x2, y2 = closest_box
                center_x = (x1 + x2) / 2
                # print(f"中心位置：{center_x}, 顶部位置：{y1}")

                # 速度决策
                allowable_deviation = 90.0
                diff = center_x - 320.0
                if diff <= -allowable_deviation:
                    angular_z = 0.07
                    linear_x = 0.0
                elif diff >= allowable_deviation:
                    angular_z = -0.07
                    linear_x = 0.0
                else:
                    angular_z = 0.0

                    # 向前移动决策
                    if y1 > 10:
                        linear_x = 0.15
                    else:
                        linear_x = 0.0

                # 更新速度
                cmd_x_z = [linear_x, angular_z]

            # 发布速度 
            velocity_publisher.publish_velocity(cmd_x_z[0], cmd_x_z[1])

            count += 1
            elapsed_time = time.time() - start_time  # 计算运行时间
            
            if elapsed_time >= 1.0:  # 每秒打印一次
                frequency = count / elapsed_time
                print(f"循环频率: {frequency:.2f} 次/秒")
                count = 0  # 重置计数
                start_time = time.time()  # 重新计时

            time.sleep(0.003)
                
    except KeyboardInterrupt:
        print("停止检测线程...")
        detection_thread.stop()  # 先停止线程
        # detection_thread.join()  # 等待线程退出
        
        print("停止机器人...")
        velocity_publisher.publish_velocity(0.0, 0.0)  # 停止小车
        time.sleep(0.2)

        # rclpy.shutdown()

    finally:
        velocity_publisher.publish_velocity(0.0, 0.0)  # 停止小车
        time.sleep(0.2)
        rclpy.shutdown()
  


if __name__ == '__main__':
    main()