import rclpy
from rclpy.node import Node
from std_msgs.msg import Int32MultiArray
from geometry_msgs.msg import Twist
from nav_msgs.msg import Odometry
from std_msgs.msg import Int32

import tf_transformations
import threading
import time
import math
from rclpy.executors import MultiThreadedExecutor

# ours
# 大模型
from spacemit_llm import LLMModel, FCModel
llm_model = LLMModel()
fc_model = FCModel(fc_model_name = 'qwen2.5-0.5b-agv-fc')

# 音频
from spacemit_audio import play_wav_non_blocking, play_wav, play_audio_in_thread
from spacemit_audio import ASRModel, RecAudio, RecAudioThread, RecAudioThreadPipeLine
asr_model = ASRModel()
from tools.agv import FuncControllerNode

record_device = 3
play_device = 'plughw:2,0'
max_time_record = 2

class AngleSubscriber(Node):
    def __init__(self):
        super().__init__('angle_subscriber')
        self.subscription = self.create_subscription(
            Int32MultiArray,
            'angle_topic',
            self.listener_callback,
            10
        )
        self.angle = 0
        self.lock = threading.Lock() # 锁，用于线程数据同步
        self.trigger_event = threading.Event()

    def listener_callback(self, msg):
        with self.lock:
            if msg.data and len(msg.data) >= 2:
                self.angle = msg.data[0]
                if msg.data[1] == 1:
                    print("Trigger received! Angle:", self.angle)
                    self.trigger_event.set()

class OdomSubscriber(Node):
    def __init__(self):
        super().__init__('odom_subscriber')
        # 订阅里程计（用于获取当前朝向）
        self.odom_subscription = self.create_subscription(
            Odometry,
            'odom',
            self.odom_callback,
            10
        )
        self.current_yaw = 0.0  # 当前朝向
        self.lock = threading.Lock() 

    def odom_callback(self, msg):
        """解析里程计数据，更新当前机器人朝向"""
        _q = msg.pose.pose.orientation
        _, _, yaw = tf_transformations.euler_from_quaternion([_q.x, _q.y, _q.z, _q.w])
        # print(f"里程计角度: {math.degrees(self.current_yaw) + 180}")
        with self.lock:
            self.current_yaw = yaw

class RotateRobot(Node):
    def __init__(self, odom_subscriber, angle_subscriber):
        super().__init__('rotate_robot')

        self.odom_subscriber = odom_subscriber  # 里程计订阅节点
        self.angle_subscriber = angle_subscriber  # 角度订阅节点
        self.cmd_vel_publisher = self.create_publisher(Twist, 'cmd_vel', 20)

    def rotate_to_angle(self):
        """让机器人旋转到目标角度"""
        with self.angle_subscriber.lock:
            audio_angle = self.angle_subscriber.angle  # 获取目标角度

        if audio_angle is None or audio_angle == 0 or audio_angle == 360:
            return

        with self.odom_subscriber.lock:
            odom_start_raw = math.degrees(self.odom_subscriber.current_yaw) + 180  # 获取当前朝向, 转换到 0 - 360 度

         # 计算旋转方向和目标角度
        if 0 < audio_angle < 180:  # 顺时针旋转
            direction = -1.0
            rotate_angle = audio_angle
        else:  # 逆时针旋转
            direction = 1.0
            rotate_angle = 360 - audio_angle

        # 里程计目标角度计算
        if direction < 0:
            odom_target_raw = odom_start_raw - rotate_angle
            if odom_target_raw < 0:
                odom_target_raw += 360
        else:
            odom_target_raw = odom_start_raw + rotate_angle
            if odom_target_raw > 360:
                odom_target_raw -= 360

        print(f"要旋转: {rotate_angle}°，方向: {direction}，当前角度: {odom_start_raw}°，目标角度: {odom_target_raw}°")

        full_error = odom_start_raw - odom_target_raw
        if full_error < 0:
            signed_flag = -1.0
        else:
            signed_flag = 1.0

        
        # circle_time = 14.5 # 一圈的时间
        # rotate_time = circle_time * rotate_angle / 360.0
        # print(f"旋转时间: {rotate_time}")

        speed_base = 0.2
        twist = Twist()
        twist.angular.z = speed_base * direction
        self.cmd_vel_publisher.publish(twist)

        # time.sleep(rotate_time)

        while True:
            with self.odom_subscriber.lock:
                current_yaw = math.degrees(self.odom_subscriber.current_yaw) + 180  # 更新当前里程计角度

            error = current_yaw - odom_target_raw

            print(error)
            if abs(error) < 6:
                self.cmd_vel_publisher.publish(Twist())
                break

            time.sleep(0.08)
        # 停止机器人
        for i in range(0, 3):
            self.cmd_vel_publisher.publish(Twist())
        self.get_logger().info("Rotation complete.")

    def stop_move(self):
        # 停止机器人
        self.cmd_vel_publisher.publish(Twist())

def run_executor(executor):
    executor.spin()  # 运行 ROS 2 事件循环（不会阻塞主线程）



def main():
    rclpy.init()

    func_controller_node = FuncControllerNode() # 总控制
    angle_sub_node = AngleSubscriber()  # 声源定位信息订阅
    odom_sub_node = OdomSubscriber()    # 里程计信息订阅
    node_rotate = RotateRobot(odom_sub_node, angle_sub_node)  # 传递订阅节点的引用

    # 多线程执行的管理器
    executor = MultiThreadedExecutor()
    executor.add_node(func_controller_node)
    executor.add_node(angle_sub_node)
    executor.add_node(odom_sub_node)
    executor.add_node(node_rotate)

    executor_thread = threading.Thread(target=run_executor, args=(executor,), daemon=True)
    executor_thread.start()

    # 录音功能类初始化
    rec_audio = RecAudioThreadPipeLine(vad_mode=1, sld=2, max_time=2, channels=1, rate=16000, device_index=record_device)

    try:
        while rclpy.ok():
            print("等待唤醒...")
            angle_sub_node.trigger_event.wait()
            angle_sub_node.trigger_event.clear()

            # 先停止小车跟随
            func_controller_node.robot_stop_move()

            node_rotate.rotate_to_angle()  # 触发旋转

            # 启动录音
            rec_audio.frame_is_append = False
            rec_audio.max_time_record = 3
            rec_audio.start_recording()

            # 播放唤醒反馈
            wav_file_path = "tools/feedback_voice/awake.wav"
            play_wav(wav_file_path, device=play_device)

            # 开始录制用户声音
            time.sleep(0.1) # 消除回音
            rec_audio.max_time_record = max_time_record
            rec_audio.frame_is_append = True
            rec_audio.time_start = time.time()

            rec_audio.thread.join() # 等待录音完成
            audio_ret = rec_audio.get_audio_file() # 获取录音文件路径

            # 语音转文字
            text = asr_model.generate(audio_ret)
            print('user: ', text)

            # 模糊匹配
            ret, func_name = func_controller_node.execute_command(text)
            if ret:
                print(f"调用的函数名: {func_name}")
                func_controller_node.function_dict[func_name]()
                continue 

            else:
                # wav_file_path = "tools/feedback_voice/search_voice.wav"
                # play_wav(wav_file_path, device=play_device)
                continue

            # 大模型推理
            t1 = time.time()
            func_map = func_controller_node.function_dict
            function_called, func_name, args = fc_model.get_function_name(text, func_map)
            print('used time:', time.time() - t1)
            if function_called:
                print(f"调用的函数名: {func_name}, 函数参数：{args}")
                func_controller_node.function_dict[func_name]()
                time.sleep(0.2)
                continue

            # llm_output = llm_model.generate(text)
            # for output_text in llm_output:
            #     print(output_text, end='', flush=True)
         

    except KeyboardInterrupt:
        print("程序终止")

    finally:
        node_rotate.stop_move()
        time.sleep(0.1)
        executor.shutdown()
        if rclpy.ok():  # 只有在 rclpy 仍然运行时才调用 shutdown
            rclpy.shutdown()

if __name__ == '__main__':
    main()
